# 計算基数を実装
import numpy as np
import sys
sys.path.append('./../')
from amaryllis.util.interest import *


def D(i, population, x):
    """
    Dxを実装

    Parameters
    ----------
    i: float
        年利率
    population: int
        x年における生存数
    x: int
        年

    Returns
    -------
    D: float
        Dxの計算基数
    """

    v = present_value_rate(i)
    D = (v ** x) * population

    return D


def N(i, population, x, lifespan=112):
    """
    Nxを実装
    x年から最終年齢までの、Dxの和

    Parameters
    ----------
    i: float
        年利率
    population: ndarray<int>
        生存人口　
    x: int
        最初の年
    lifespan: int, default 112
        最終年齢

    Returns
    -------
    N: float
        Nxの計算基数  
    """

    assert (population.shape[0] == lifespan - x + 1) and (len(population.shape) == 1), "生存人口の入力数が異なります"

    Dxs = [D(i, p, y) for y, p in zip(range(x, lifespan + 1), population)]
    Dxs = np.array(Dxs)

    N = np.sum(Dxs)

    return N


def C(i, dead_population, x):
    """
    Cxの計算基数を実装

    Parameters
    ----------
    i: float
        年利率
    dead_population: int
        x年度死亡人口
    x: int
        最初の年

    Returns
    -------
    C: float
        Cxの計算基数
    """

    v = present_value_rate(i)
    C = v ** (x + 1) * dead_population

    return C


def M(i, dead_population, x, lifespan=112):
    """
    Mの計算基数を計算
    Cxの最終年齢までの和

    Parameters
    ----------
    i: float
        年利率
    dead_population: ndarray<int>
        x~lifespan年度死亡人口
        size: (lifespan - x, )
    x: int
        最初の年度
    lifespan: int, default 112
        最終年齢

    Returns
    -------
    M: float
        Mxの計算基数
    """

    assert (dead_population.shape[0] == lifespan - x + 1) and (len(dead_population.shape) == 1), "死亡人口の入力が異なります"

    Cxs = [C(i, p, y) for y, p in zip(range(x, lifespan + 1), dead_population)]
    Cxs = np.array(Cxs)

    M = np.sum(Cxs)

    return M


def S(i, population, x, lifespan=112):
    """
    Sxの計算基数を実装
    Mxの最終年齢までの和

    Parameters
    ----------
    i: float
        年利率
    population: ndarray<int>
        x~lifespan年度生存人口
        size: (lifespan - X, )
    x: int
        最初の年
    lifespan: int, default 120
        最終年齢

    Returns
    -------
    S: float
        Sxの計算基数
    """

    assert (population.shape[0] == lifespan - x + 1) and (len(population.shape) == 1), "生存人口の入力数が異なります"

    S = 0

    for j in range(0, lifespan - x):
        S += (j + 1) * D(i, population[j], j)

    return S


def R(i, dead_population, x, lifespan=112):
    """
    Rの計算基数を実装
    Mxの最終年齢までの和

    Parameters
    ----------
    i: float
        年利率
    dead_population: ndarray<int>
        x~lifespan年度までの死亡人口
    x: int
        最初の年
    lifespan: int, default 120
        最終年齢

    Returns
    -------
    R: float
        Rの計算基数
    """

    assert (dead_population.shape[0] == lifespan - x + 1) and (len(dead_population.shape) == 1), "死亡人口の入力が異なります"

    R = 0
    for j in range(0, lifespan - x):
        R += (j + 1) * C(i, dead_population[j], j)

    return R


def C_continuous(i, dead_population, x):
    """
    C_の計算基数を実装

    Parameters
    ----------
    i: float
        年利率
    dead_population: int
        死亡人口
    x: int
        年度

    Returns
    -------
    C: float
        C_の計算基数
    """

    v = present_value_rate(i)
    C = v ** (x + 1 / 2) * dead_population

    return C


def M_continuous(i, dead_population, x, lifespan=112):
    """
    M_の計算基数を実装

    Parameters
    ----------
    i: float
        年利率
    dead_population: ndarray<int>
        x~lifespan年度における死亡人口
    x: int
        最初の年度
    lifespan: int, default 120
        最終年齢

    Returns
    -------
    M: float
        M_の計算基数
    """

    assert (dead_population.shape[0] == lifespan - x + 1) and (len(dead_population.shape) == 1), "死亡人口の入力が異なります"

    Cxs = [C_continuous(i, p, y) for y, p in zip(
        range(x, lifespan + 1), dead_population)]
    Cxs = np.array(Cxs)

    M = np.sum(Cxs)

    return M



# 生存人口、死亡人口の変換を実装

import numpy as np


def living_to_dead(population, lifespan=112):
    """
    生存人口を死亡人口に変換

    Parameters
    ----------
    population: ndarray<int>
        生存人口
        size: (lifespan, )
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    dead_population: ndarray<int>
        死亡人口
        size: (lifespan, )
    """

    assert len(population.shape) == 1, '生存人口の入力が異なります'
    assert population.shape[0] == lifespan + 1, '生存人口は0歳から最終年齢分まで入力してください'

    dead_population = []

    for j in range(0, lifespan):
        dead_population.append(population[j] - population[j + 1])
    
    dead_population.append(population[lifespan])

    dead_population = np.array(dead_population)

    return dead_population


# 利息計算を行う
import numpy as np


def present_value_rate(i=0.03):
    """
    現価率を計算

    Parameters
    ----------
    i: float, default 0.03
        年利率

    Returns
    -------
    v: float
        現価率
    """

    v = 1 / (1 + i)

    return v


def discount_rate(i=0.03):
    """
    割引率を計算

    Parameters
    ----------
    i: float, default 0.03
        年利率

    Returns
    -------
    d: float
        割引率
    """

    d = i / (1 + i)

    return d


def name_interest_rate(i=0.03, k=1):
    """
    名称利率を計算

    Parameters
    ----------
    i: float, default 0.03
        年利率
    k: int, default 1
        一年に支払う回数

    Returns
    -------
    i_k: float
        年k回払いの名称利率
    """

    i_k = k * ((1 + i) ** (1 / k) - 1)

    return i_k


def name_discount_rate(i=0.03, k=1):
    """
    名称割引率を計算

    Parameters
    ----------
    i:　float, default 0.03
        年利率
    k: int, default 1
        一年に支払う回数

    Returns
    -------
    d_k: float
        年k回払いの名称割引率
    """

    i_k = name_interest_rate(i, k)

    d_k = i_k / (1 + i_k / k)

    return d_k


def force_of_interest(i=0.03):
    """
    利力を計算

    Parameters
    ----------
    i: float, default 0.03
        年利率

    Returns
    -------
    delta: float
        利力
    """

    delta = np.log(1 + i)

    return delta



# 生命年金モデルを実装

import numpy as np
import sys
sys.path.append('./../../')
from amaryllis.util.interest import *
from amaryllis.util.calc_base import *
from amaryllis.config import *


def life_annuity(i=I, population=POPULATION, x=X, n=N, due=True, f=F, lifespan=LIFESPAN):
    """
    生命年金の原価を算出する(a[x : n])

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
        size: (lifespan, )
    x: int, default 30
        契約時年齢
    n: int, default 20
        払込期間
    due: bool, default True
        trueの場合、期始払
        falseの場合、期末払
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢

    Returns
    -------
    a: float
        年金現価
    """
    assert (population.shape[0] == lifespan + 1) and (len(population.shape) == 1), '生存人口の入力が異なります'
    
    Dx = D(i, population[x], x)
    if due:  
        assert x + n + f <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
        Nx = N(i, population[x + f :], x + f, lifespan)
        Nx_n = N(i, population[x + n + f :], x + n + f, lifespan)
        a = (Nx - Nx_n) / Dx

        return a

    else:
        assert x + n + f + 1 <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
        Nx_1 = N(i, population[x + f + 1 :], x + f + 1, lifespan)
        Nx_n1 = N(i, population[x + n + f + 1 :], x + n + f + 1, lifespan)
        a = (Nx_1 - Nx_n1) / Dx

        return a
      

def lifelong_annuity(i=I, population=POPULATION, x=X, due=True, f=F, lifespan=LIFESPAN):
    """
    終身生命年金の現価を算出(a[x])

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
        size: (lifespan, )
    x: int, default 30
        契約時の年齢
    due: bool, default True
        Trueなら期始払、Falseなら期末払
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    a: float
        終身生命年金現価
    """

    assert (population.shape[0] == lifespan + 1) and (len(population.shape) == 1), '生存人口の入力が異なります'

    Dx = D(i, population[x], x)
    if due:
        assert x + f <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
        Nx = N(i, population[x + f :], x + f, lifespan)
        a = Nx / Dx

        return a

    else:
        assert x + f + 1 <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
        Nx_1 = N(i, population[x + f + 1 :], x + f + 1, lifespan)
        a = Nx_1 / Dx

        return a
            

def increasing_life_annuity(i=I, population=POPULATION, x=X, n=N, due=True, f=F, lifespan=LIFESPAN):
    """
    累加生命年金現価を算出(Ia[x : n])

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
    x: int, default 30
        契約時年齢
    n: int, default 20
        払込期間
    due: bool, default True
        Trueなら期始払、Falseなら期末払
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    Ia: float
        年金現価
    """

    assert (population.shape[0] == lifespan + 1) and (len(population.shape) == 1), '生存人口の入力が異なります'

    Dx = D(i, population[x], x)
    if due:
        assert x + f + n <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
        Sx = S(i, population[x + f :], x + f, lifespan)
        Sx_n = S(i, population[x + f + n :], x + f + n, lifespan)
        Nx_n = N(i, population[x + f + n :], x + f + n, lifespan)
        Ia = (Sx - Sx_n - n * Nx_n) / Dx

        return Ia
    
    else:
        assert x + f + n + 1 <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
        Sx_1 = S(i, population[x + f + 1 :], x + f + 1, lifespan)
        Sx_n1 = S(i, population[x + f + n + 1 :], x + f + n + 1, lifespan)
        Nx_n1 = N(i, population[x + f + n + 1 :], x + f + n + 1, lifespan)
        Ia = (Sx_1 - Sx_n1 - n * Nx_n1) / Dx

        return Ia


def increasing_lifelong_annuity(i=I, population=POPULATION, x=X, n=N, due=True, f=F, lifespan=LIFESPAN):
    """
    累加終身生命年金現価を算出(Ia[x])

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
        size: (lifespan, )
    x: int, default 30
        契約年齢
    n: int, default 20
        年金額増加回数上限
    due: bool
        Trueなら期始払、Falseなら期末払
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    Ia: float
        年金現価
    """

    assert (population.shape[0] == lifespan + 1) and (len(population.shape) == 1), '生存人口の入力が異なります'

    Dx = D(i, population[x], x)
    if due:
        assert x + f + n <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
        Sx = S(i, population[x + f :], x + f, lifespan)
        Sx_n = S(i, population[x + f + n :], x + f + n, lifespan)
        Ia = (Sx - Sx_n) / Dx

        return Ia
    
    else:
        assert x + f + n + 1 <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
        Sx_1 = S(i, population[x + f + 1 :], x + f + 1, lifespan)
        Sx_n1 = S(i, population[x + f + n + 1 :], x + f + n + 1, lifespan)
        Ia = (Sx_1 - Sx_n1) / Dx

        return Ia


def decreasing_life_annuity(i=I, population=POPULATION, x=X, n=N , due=True, f=F, lifespan=LIFESPAN):
    """
    累減生命年金現価を算出(Ia[x : n])

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
    x: int, default 30
        契約時年齢
    n: int, default 20
        払込期間
    due: bool, default True
        Trueなら期始払、Falseなら期末払
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢

    Returns
    -------
    Ia: float
        年金現価
    """

    assert (population.shape[0] == lifespan + 1) and (len(population.shape) == 1), '生存人口の入力が異なります'

    Dx = D(i, population[x], x)
    if due:
        assert x + f + n + 1 <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
        Sx_1 = S(i, population[x + f + 1 :], x + f + 1, lifespan)
        Sx_n1 = S(i, population[x + f + n + 1 :], x + f + n + 1, lifespan)
        Nx = N(i, population[x + f:], x + f, lifespan)
        Ia = (n * Nx - Sx_1 + Sx_n1) / Dx

        return Ia
    
    else:
        assert x + f + n + 2 <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
        Sx_2 = S(i, population[x + f + 2 :], x + f + 2, lifespan)
        Sx_n2 = S(i, population[x + f + n + 2 :], x + f + n + 2, lifespan)
        Nx_1 = N(i, population[x + f + 1 :], x + f + 1, lifespan)
        Ia = (n * Nx_1 - Sx_2 + Sx_n2) / Dx
        
        return Ia


# 一時払保険料現価

import numpy as np
import sys
sys.path.append('./../')
from amaryllis.util.calc_base import *
from amaryllis.util.interest import *
from amaryllis.survival_rate.population import *
from amaryllis.config import *


def single_term_insurance(i=I, population=POPULATION, x=X, n=N, f=F, lifespan=LIFESPAN):
    """
    定期保険の一時払保険料現価を算出

    Parameters
    ---------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
        size: (lifespan, )
    x: int, default 30
        契約年齢
    n: int, default 20
        契約年数
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    A: float
        保険料現価
    """

    assert (population.shape[0] == lifespan + 1) and (len(population.shape) == 1), '生存人口の入力が異なります'
    assert x + f + n <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
    dead_population = living_to_dead(population, lifespan)
    Dx = D(i, population[x], x)
    
    Mx = M(i, dead_population[x + f :], x + f, lifespan)
    Mx_n = M(i, dead_population[x + f + n :], x + f + n, lifespan)

    A = (Mx - Mx_n) / Dx

    return A


def single_life_insurance(i=I, population=POPULATION, x=X, n=N, f=F, lifespan=LIFESPAN):
    """
    生存保険の一時払い保険料現価を算出

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口　
    x: int, default 30
        契約年齢
    n: int, default 20
        契約年数
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    A: float
        保険料現価
    """

    assert (population.shape[0] == lifespan + 1) and (len(population.shape) == 1), '生存人口の入力が異なります'
    assert x + f + n <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
    Dx = D(i, population[x], x)
    Dx_n = D(i, population[x + f + n], x + f + n)

    A = Dx / Dx

    return A


def single_endowment_insurance(i=I, population=POPULATION, x=X, n=N, f=F, lifespan=LIFESPAN):
    """
    養老保険の一時払保険現価を算出

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
        size: (lifespan, )
    x: int, default 30
        契約年齢
    n: int, default 20
        契約年数
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    A: float
        保険料現価
    """

    A = single_term_insurance(i, population, x, n, f, lifespan) + single_life_insurance(i, population, x, n, f, lifespan)

    return A


def single_increasing_term_insurance(i=I, population=POPULATION, x=X, n=N, f=F, lifespan=LIFESPAN):
    """
    累加定期保険の一時払い保険料現価を算出

    Parameters
    ----------
    i: float, default I
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
        size: (lifespan, )
    x: int, default 30
        契約年齢
    n: int, default 20
        契約年数
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    IA: float
        保険料現価
    """

    assert (len(population.shape) == 1) and (population.shape[0] == lifespan + 1), '生存人口の入力が異なります'
    assert x + f + n <= lifespan, '契約終了前に最終年齢を迎えてしまいます'

    dead_population = living_to_dead(population, lifespan)
    Dx = D(i, population[x], x)
    Rx = R(i, dead_population[x + f :], x + f, lifespan)
    Rx_n = R(i, dead_population[x + f + n :], x + f + n, lifespan)
    Mx_n = M(i, dead_population[x + f + n :], x + f + n, lifespan)

    IA = (Rx - Rx_n - n * Mx_n) / Dx

    return IA


def single_decreasing_term_insurance(i=I, population=POPULATION, x=X, n=N, f=F, lifespan=LIFESPAN):
    """
    逓減定期保険一時払いの保険料現価を算出

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口　
        size: (lifespan, )
    x: int, default 30
        契約年齢
    n: int, default 20
        契約年数
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    DA: float
        保険料現価
    """
    
    assert (len(population.shape) == 1) and (population.shape[0] == lifespan + 1), '生存人口の入力が異なります'
    assert x + f + n + 1 <= lifespan, '契約終了前に最終年齢を迎えてしまいます'

    dead_population = living_to_dead(population, lifespan)
    Dx = D(i, population[x], x)
    Mx = M(i, dead_population[x + f :], x + f, lifespan)
    Rx_1 = R(i, dead_population[x + f + 1 :], x + f + 1, lifespan)
    Rx_n1 = R(i, dead_population[x + f + n + 1 :], x + f + n + 1, lifespan)

    DA = (n * Mx - Rx_1 + Rx_n1) / Dx

    return DA


def single_lifelong_insurance(i=I, population=POPULATION, x=X, f=F, lifespan=LIFESPAN):
    """
    終身保険の一時払い保険料現価を算出

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
        size: (lifespan, )
    x: int, default 30
        契約年齢
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    A: float
        保険料現価
    """

    assert (len(population.shape) == 1) and (population.shape[0] == lifespan + 1), '生存人口の入力が異なります'
    assert x + f <= lifespan, '契約終了前に最終年齢を迎えてしまいます'

    dead_population = living_to_dead(population, lifespan)

    Dx = D(i, population[x], x)
    Mx = M(i, dead_population[x + f :], x + f, lifespan)

    A = Mx / Dx

    return A


def single_increasing_lifelong_insurance(i=I, population=POPULATION, x=X, n=N, f=F, lifespan=LIFESPAN):
    """
    累加終身保険一時払い保険料現価を算出

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口　
        size: (lifespan, )
    x: int, default 30
        契約年齢
    n: int, default 20
        保険料が増大する回数の上限
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    IA: float
        保険料現価
    """

    assert (len(population.shape) == 1) and (population.shape[0] == lifespan + 1), '生存人口の入力が異なります'
    assert x + f + n <= lifespan, '契約終了前に最終年齢を迎えてしまいます'
    
    dead_population = living_to_dead(population, lifespan)
    Dx = D(i, population[x], x)
    Rx = R(i, dead_population[x + f :], x + f, lifespan)
    Rx_n = R(i, dead_population[x + f + n :], x + f + n, lifespan)

    IA = (Rx - Rx_n) / Dx

    return IA


def single_decreasing_lifelong_insurance(i=I, population=POPULATION, x=X, n=N, f=F, lifespan=LIFESPAN):
    """
    逓減終身保険の一時払い保険料現価を算出

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION 
        生存人口
        size: (lifespan, )
    x: int, default 30
        契約年齢
    n: int, default 20
        保険料減少回数の上限
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    DA: float
        保険料現価
    """

    assert (len(population.shape) == 1) and (population.shape[0] == lifespan + 1), '生存人口の入力が異なります'
    assert x + f + n <= lifespan, '契約終了前に最終年齢を迎えてしまいます'

    dead_population = living_to_dead(population, lifespan)
    Dx = D(i, population[x], x)
    Mx = M(i, dead_population[x + f :], x + f, lifespan)
    Rx_1 = R(i, dead_population[x + 1 :], x + 1, lifespan)
    Rx_n = R(i, dead_population[x + f + n :], x + f + n, lifespan)

    DA = (n * Mx - Rx_1 + Rx_n) / Dx

    return DA


def single_continuous_term_insurance(i=I, population=POPULATION, x=X, n=N, f=F, lifespan=LIFESPAN):
    """
    即時払い定期保険の一時払い保険料現価を算出

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
        size: (lifespan, )
    x: int, default 30
        契約年齢
    n: int, default 20
        契約年数
    f: int, default 10
        据置年数
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    A: float
        保険料現価
    """

    assert (len(population.shape) == 1) and (population.shape[0] == lifespan + 1), '生存人口の入力が異なります'
    assert x + f + n <= lifespan, '契約終了前に最終年齢を迎えてしまいます'

    dead_population = living_to_dead(population, lifespan)
    Dx = D(i, population[x], x)

    Mx = M_continuous(i, dead_population[x + f :], x + f, lifespan)
    Mx_n = M_continuous(i, dead_population[x + f + n :], x + f + n, lifespan)

    A = (Mx - Mx_n) / Dx

    return A


def single_continuous_lifelong_insurance(i=I, population=POPULATION, x=X, f=F, lifespan=LIFESPAN):
    """
    即時払い終身保険の一時払い保険料現価を算出

    Parameters
    ----------
    i: float, default 0.03
        年利率
    population: ndarray<int>, default POPULATION
        生存人口
        size: (lifespan, )
    x: int, default 30
        契約年齢
    f: int, default 10
        据置期間
    lifespan: int, default 112
        最終年齢
    
    Returns
    -------
    A: float
        保険料現価
    """

    assert (len(population.shape) == 1) and (population.shape[0] == lifespan + 1), '生存人口の入力が異なります'
    assert x + f <= lifespan, '契約終了前に最終年齢を迎えてしまいます'

    dead_population = living_to_dead(population, lifespan)
    Dx = D(i, population[x], x)

    Mx = M_continuous(i, dead_population[x + f :], x + f, lifespan)

    A = Mx / Dx

    return A





    